//******************************************************************
//* uDRIVE API
//* LAST UPDATED 03/17/2008
//* FIRMWARE BY FRED EADY FOR SERVO MAGAZINE
//* CHANGES/ADDITIONS
//******************************************************************
#include <pic18.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <htc.h>
#include <stdlib.h>

// Settings for PIC18F2620
__CONFIG(1,IESODIS & FCMDIS & HS);
__CONFIG(2,BORDIS & PWRTDIS & WDTDIS);
__CONFIG(3,MCLREN & LPT1DIS & PBDIGITAL);
__CONFIG(4,DEBUGEN & LVPDIS & XINSTDIS & STVREN);
__CONFIG(5,UNPROTECT);
__CONFIG(6,WRTEN);
__IDLOC(0317);

//******************************************************************
//*	FUNCTION PROTOTYPES
//******************************************************************
void interrupt TIMERS(void);
void putch(char c);
unsigned int recvchar(void);
unsigned int sendchar(unsigned int data);
char CharInQueue(void);
void init(void);
void initEEUSART(void);
char init_disk(void);
char set_mem_addr(unsigned long addr);
//******************************************************************
//*	PIC18F2620 RAM Definitions
//******************************************************************
char sector_buffer[512],read_buffer[512],rc;
unsigned int index;
unsigned int temp16,scratch16;
char temp8,scratch8,bytein;
char hundreds, tens, ones;
//t = optional timer value 
//i = interrupt counter value 
unsigned int tmsecs1,tmsecs2,tmsecs3;
char tsecs1,tsecs2,tsecs3,tmins1,tmins2,tmins3,thours1,thours2,thours3;
unsigned int imsecs1,imsecs2,imsecs3;
char isecs1,isecs2,isecs3,imins1,imins2,imins3,ihours1,ihours2,ihours3;
char hours, mins, secs;
//******************************************************************
//    MACROS AND PORT/PIN DEFINITIONS
//******************************************************************
#define bitset(var, bitno) ((var) |= 1 << (bitno))
#define bitclr(var, bitno) ((var) &= ~(1 << (bitno)))
#define make8(var,offset)	((unsigned int)var >> (offset * 8)) & 0x00FF
#define	make16(varhigh,varlow)	(((unsigned int)varhigh & 0xFF)* 0x100) + ((unsigned int)varlow & 0x00FF)
#define make32(var1,var2,var3,var4) \
		((unsigned long)var1<<24)+((unsigned long)var2<<16)+ \
		((unsigned long)var3<<8)+((unsigned long)var4)		
#define make32i(var1,var2) ((unsigned long)var1<<16)+((unsigned long)var2)
//******************************************************************
//*	GENERAL DEFINITONS
//******************************************************************
#define	act_led	LATA5
//******************************************************************
//*	uDRIVE DEFINITONS
//******************************************************************
#define uRESET	LATC5
#define uTX		LATC6
#define uACK	0x06
#define uNAK	0x15
//******************************************************************
//*	INTERRUPT DEFINITIONS
//******************************************************************
#define		disable_RXint	  	RCIE = 0
#define		enable_RXint	  	RCIE = 1
#define		disable_TMR2int  	TMR2IE = 0
#define		enable_TMR2int	 	TMR2IE = 1
#define		disable_TMR3int  	TMR3IE = 0
#define		enable_TMR3int	 	TMR3IE = 1
#define		disable_TMR1int  	TMR1IE = 0
#define		enable_TMR1int	 	TMR1IE = 1

#define		enable_GLOBALint  	IPEN = 0;				\
							  	GIE = 1;				\
							  	PEIE = 1;
#define		disable_GLOBALint 	IPEN = 0;				\
							  	GIE = 0;				\
							  	PEIE = 0;
#define		TIMER0ON			TMR0ON = 1
#define		TIMER0OFF			TMR0ON = 0
#define		TIMER1ON			TMR1ON = 1
#define		TIMER1OFF			TMR1ON = 0
#define		TIMER2ON		  	TMR2ON = 1
#define		TIMER2OFF		  	TMR2ON = 0
#define		TIMER3ON		  	TMR3ON = 1
#define		TIMER3OFF		  	TMR3ON = 0

#define mdelay1(msecdelay)		TIMER1OFF;					\
								TMR1IF = 0;					\
								imsecs1 = 0;				\
								tmsecs1 = 0;				\
								TIMER1ON;					\
								while(tmsecs1 < msecdelay);
#define	sdelay1(secdelay)		TIMER1OFF;					\
								TMR1IF = 0;					\
								imsecs1 = 0;				\
								tmsecs1 = 0;				\
								isecs1 = 0;					\
								tsecs1 = 0;					\
								TIMER1ON;					\
								while(tsecs1 < secdelay);

//******************************************************************
//*	EUSART BUFFER DEFINITIONS
//******************************************************************
#define BAUD	57600		//desired baud rate
#define FOSC	20000000	//oscillator frequency
#define DIVIDER ((unsigned int)(FOSC/(16 * BAUD) - 1))
//1,2,4,8,16,32,64,128 or 256 bytes 
#define EUSART_RX_BUFFER_SIZE  256     
#define EUSART_RX_BUFFER_MASK ( EUSART_RX_BUFFER_SIZE - 1 )
// 1,2,4,8,16,32,64,128 or 256 bytes 
#define EUSART_TX_BUFFER_SIZE  256      
#define EUSART_TX_BUFFER_MASK ( EUSART_TX_BUFFER_SIZE - 1 )
unsigned char EUSART_RxBuf[EUSART_RX_BUFFER_SIZE],EUSART_TxBuf[EUSART_TX_BUFFER_SIZE];
unsigned char EUSART_TxHead,EUSART_TxTail,EUSART_RxHead,EUSART_RxTail;

void putch(char c)
{
   while(!TRMT) 	// TRMT is set when TSR is empty
      continue;

  TXREG = c;		// load the register
}


unsigned int recvchar(void)
{
  unsigned char tmptail;
                                // wait for incomming usart_data  
  while ( EUSART_RxHead == EUSART_RxTail );
                                // calculate buffer index  
  tmptail = ( EUSART_RxTail + 1 ) & EUSART_RX_BUFFER_MASK;
  EUSART_RxTail = tmptail;        // store new index  

  return EUSART_RxBuf[tmptail];   // return usart_data  
}

unsigned int sendchar(unsigned int usart_data)
{
  unsigned char tmphead;
                                // calculate buffer index 
  tmphead = ( EUSART_TxHead + 1 ) & EUSART_TX_BUFFER_MASK;
                                // wait for free space in buffer 
  while ( tmphead == EUSART_TxTail );
                                // store usart_data in buffer 
  EUSART_TxBuf[tmphead] = (unsigned char)usart_data;
  EUSART_TxHead = tmphead;        // store new index 

  TXIE = 1;            // enable TX interrupt 

  return usart_data;
}

unsigned char CharInQueue(void)
{
  return(EUSART_RxHead != EUSART_RxTail);
}

//******************************************************************
//*	Init EUSART Function
//******************************************************************
void init_EEUSART(void)
{
  SPBRG = DIVIDER;	//load baud rate divisor-calc didn't work here	
  TRISC7 = 1;			//receive pin
  TRISC6 = 0;			//transmit pin
  TXSTA = 0x04;		//high speed baud rate	
  RCSTA = 0x80;		//enable serial port and serial port pins
  
  EUSART_RxTail = 0x00;	//flush receive buffer 
  EUSART_RxHead = 0x00;
  EUSART_TxTail = 0x00;	//flush transmit buffer
  EUSART_TxHead = 0x00;
  RCIP = 1;			//receive interrupt = high priority
  TXIP = 1;			//transmit interrupt = high priority
  RCIE = 1;			//enable receive interrupt
  PEIE = 1;			//enable all unmasked peripheral interrupts
  GIE = 1;			//enable all unmasked interrupts
  CREN = 1;			//enable EUSART1 receiver
  TXIE = 0;			//disable EUSART1 transmit interrupt
  TXEN = 1;			//transmitter enabled
}

//******************************************************************
//*	INTERRUPT HANDLER ROUTINE
//******************************************************************
void interrupt TIMERS(void)
{
  unsigned char usart_data,tmphead,tmptail;

	if((TMR3IF && TMR3IE))
	{
		TMR3IF = 0;
		TMR3H = 0xEC;	
		TMR3L = 0x77;

		++tmsecs3;
		if(++imsecs3 == 1000)
		{
			imsecs3 = 0;
			++isecs3;
			++tsecs3;
			act_led ^= 1;
		}
		if(isecs3 == 60)
		{
			isecs3 = 0;
			++imins3;
			++tmins3;
		}
		if(imins3 == 60)
		{
			imins3 = 0;
			++ihours3;
		}
	}

	if((TMR1IF && TMR1IE))
	{
		TMR1IF = 0;
		TMR1H = 0xEC;	
		TMR1L = 0x77;

		++tmsecs1;
		if(++imsecs1 == 1000)
		{
			imsecs1 = 0;
			++isecs1;
			++tsecs1;
		}
		if(isecs1 == 60)
		{
			isecs1 = 0;
			++imins1;
			++tmins1;
		}
		if(imins1 == 60)
		{
			imins1 = 0;
			++ihours1;
		}
	}
//SERVICE THE EEUSART
 if(RCIF)
 {
   usart_data = RCREG;                   // read the received usart_data 
                                // calculate buffer index 
   tmphead = ( EUSART_RxHead + 1 ) & EUSART_RX_BUFFER_MASK;
   EUSART_RxHead = tmphead;        // store new index 

   if ( tmphead == EUSART_RxTail )
   {
     // ERROR! Receive buffer overflow 
   }
                                
   EUSART_RxBuf[tmphead] = usart_data;   // store received usart_data in buffer 
  }

  
  if(TRMT)
  {
                                // check if all usart_data is transmitted  
   if ( EUSART_TxHead != EUSART_TxTail )
   {
                                // calculate buffer index  
     tmptail = ( EUSART_TxTail + 1 ) & EUSART_TX_BUFFER_MASK;
     EUSART_TxTail = tmptail;      // store new index  

    TXREG = EUSART_TxBuf[tmptail];  // start transmitiON  
   }
   else
   {
     TXIE = 0;         // disable TX interrupt  
   }
  }

}
//****************************************************************************
//*	INITIALIZE ROUTINE
//****************************************************************************
void init()
{
//****************************************************************************
//*	INITIALIZE CLOCK AND IO PORTS
//****************************************************************************
	TRISA = 0b11011111;
	TRISB = 0b11011111;
	TRISC = 0b10000000;
	uTX = 1;
	uRESET = 1;
//******************************************************************
//*	CONFIGURE A2D AND COMPARATORS
//******************************************************************
	ADCON1 = 0b00001111;
	ADON = 0;
	CMCON = 0x07;
//******************************************************************
//*	CONFIGURE AND START TIMER1
//* SET TO OVERFLOW EVERY 1mS
//******************************************************************
	//F831 = 8MHz
	//E0C1 = 32MHz
	//EC77 = 20MHz
	TIMER1OFF;
	T1CON = 0b00000000;
	TMR1H = 0xEC;
	TMR1L = 0x77;
	TIMER1ON;
//******************************************************************
//*	CONFIGURE AND START TIMER3
//* SET TO OVERFLOW EVERY 1mS
//******************************************************************
	TIMER3OFF;
 	ihours3 = 12;
	imins3 = 0;
	isecs3 = 0;
  	imsecs3 = 0;
	T3CON = 0x00;		
	TMR3H = 0xEC;	
	TMR3L = 0x77;
	TIMER3ON ;
//******************************************************************
//*	CONFIGURE EXTERNAL INTERRUPTS
//******************************************************************
	enable_TMR1int;
    enable_GLOBALint;
//******************************************************************
//*	RESET THE uDRIVE
//******************************************************************
	sdelay1(1);
//******************************************************************
//*	INITIALIZE EEUSART
//******************************************************************
	init_EEUSART();
	printf("U");
	do{
		++scratch16;
	}while(!(CharInQueue()));
    bytein = recvchar();
    if(bytein == uACK)
    {
	    	enable_TMR3int;
			sdelay1(5);
	}
	else
	{
		act_led = 0;
	}
}
//******************************************************************
//*	INITIALIZE DISK DRIVE
//******************************************************************
char init_disk(void)
{
	sendchar('@');
	sendchar('i');
	do{
		++scratch8;
	}while(!(CharInQueue()));
    return (bytein = recvchar());
}
//******************************************************************
//*	SET MEMORY ADDRESS
//******************************************************************
char set_mem_addr(unsigned long addr)
{
	sendchar('@');
	sendchar('A');
	sendchar((addr & 0xFF000000) >> 24);
	sendchar((addr & 0x00FF0000) >> 16);
	sendchar((addr & 0x0000FF00) >> 8);
	sendchar((addr & 0x000000FF));
	do{
		++scratch8;
	}while(!(CharInQueue()));
    return (bytein = recvchar());
}	
//******************************************************************
//*	WRITE BYTE
//******************************************************************
char write_byte(char data,unsigned long addr)
{
	set_mem_addr(addr);
	
	sendchar('@');
	sendchar('w');
	sendchar(data);
	do{
		++scratch8;
	}while(!(CharInQueue()));
    return (bytein = recvchar());
}	
//******************************************************************
//*	WRITE 512-BYTE SECTOR
//******************************************************************
char write_sector(unsigned int sector,char *buffer)
{
	char *byteptr;
	unsigned int index;
	
	byteptr = buffer;
	buffer = 0;
	sendchar('@');
	sendchar('W');
	sendchar((sector & 0x00FF0000) >> 16);
	sendchar((sector & 0x0000FF00) >> 8);
	sendchar((sector & 0x000000FF));
	for(index=0;index<512;++index)
	{
		sendchar(*byteptr++);
	}
	do{
		++scratch8;
	}while(!(CharInQueue()));
    return (bytein = recvchar());
}
//******************************************************************
//*	READ BYTE
//******************************************************************
char read_byte(unsigned long addr)
{
	set_mem_addr(addr);
	
	sendchar('@');
	sendchar('r');
	do{
		++scratch8;
	}while(!(CharInQueue()));
    return (bytein = recvchar());
}	
//******************************************************************
//*	READ 512-BYTE SECTOR
//******************************************************************
char read_sector(unsigned int sector,char *buffer)
{
	char *byteptr;
	unsigned int index;
	
	EUSART_RxTail = 0x00;	//flush receive buffer 
  	EUSART_RxHead = 0x00;
  	
	byteptr = buffer;
	buffer = 0;
	sendchar('@');
	sendchar('R');
	sendchar((sector & 0x00FF0000) >> 16);
	sendchar((sector & 0x0000FF00) >> 8);
	sendchar((sector & 0x000000FF));
	do{
		*byteptr++ = recvchar();
		}while(*byteptr < 512);
    return (bytein = recvchar());
}

//******************************************************************
//*	MAIN SERVICE LOOP
//******************************************************************
void main(void)
{
	init();
	rc = init_disk();
	
	scratch8 = 0;	
	for(scratch16 = 0;scratch16<512;++scratch16)
	{
		sector_buffer[scratch16] = scratch8++;
	}
	rc = write_sector(0x0000,sector_buffer);
	rc = read_sector(0x0000,read_buffer);	
	
	rc = write_byte(0x42,0x00000000);
	rc = read_byte(0x00000000);
	

	while(1);
}